/*
 * Driver for unmanaged DAC TDA1543 connected to SSC Atmel's controller.
 */

#include <linux/clk.h>
#include <linux/err.h>
#include <linux/dma-mapping.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/io.h>

#include <sound/initval.h>
#include <sound/control.h>
#include <sound/core.h>
#include <sound/pcm.h>

#include <linux/spi/spi.h>

#include <linux/atmel-ssc.h>

#include <mach/gpio.h>// at91 specific GPIO pins

#undef DEBUG

#define SSC_ID 0
#define DRV_NAME "ssc_sound"
#define DEVICE_NUMBER 0

struct snd_ssc {
	struct snd_card			*card;
	struct snd_pcm			*pcm;
	struct snd_pcm_substream	*substream_playback;
	int				irq;
	int				period_playback;
	unsigned int			bitrate;
	struct ssc_device		*ssc;
	/* Protect SSC registers against concurrent access. */
	spinlock_t			lock;
};

static struct snd_card* ssc_card = NULL;

#define get_chip(card) ((struct snd_ssc *)card->private_data)

static struct snd_pcm_hardware snd_ssc_playback_hw = {
	.info		= SNDRV_PCM_INFO_INTERLEAVED |
					SNDRV_PCM_INFO_BLOCK_TRANSFER,
	.formats	= SNDRV_PCM_FMTBIT_S16_LE,
	.rates		= SNDRV_PCM_RATE_CONTINUOUS,
	.rate_min	= 5512,
	.rate_max	= 176400,
	.channels_min	= 2,
	.channels_max	= 2,
	.buffer_bytes_max = 64 * 1024 - 1,
	.period_bytes_min = 512,
	.period_bytes_max = 64 * 1024 - 1,
	.periods_min	= 4,
	.periods_max	= 1024,
};

/*
 * Calculate and set bitrate and divisions.
 */
static int snd_ssc_set_bitrate(struct snd_ssc *chip, unsigned int bitrate)
{
	unsigned int ssc_rate = clk_get_rate(chip->ssc->clk);
	unsigned int ssc_div;

	/*
	 * We connect two clocks here, picking divisors so the I2S clocks
	 * out data at the same rate the DAC clocks it in ... and as close
	 * as practical to the desired target rate.
	 *
	 * The DAC master clock (MCLK) is programmable, and is either 256
	 * or (not here) 384 times the I2S output clock (BCLK).
	 */

	/* SSC clock / (bitrate * stereo * 16-bit). */
	ssc_div = ssc_rate / (bitrate * 2 * 16);

	/* ssc_div must be a power of 2. */
	ssc_div = (ssc_div + 1) & ~1UL;
	if( ssc_div < 2 ) ssc_div = 2;
	else if( ssc_div > 8190 ) ssc_div = 8190;

	/* Set divider in SSC device. */
	ssc_writel(chip->ssc->regs, CMR, ssc_div/2);
#ifdef DEBUG
	/* SSC clock / (ssc divider * 16-bit * stereo). */
	chip->bitrate = ssc_rate / (ssc_div * 16 * 2);

	printk(KERN_INFO DRV_NAME": requested bitrate: %u, supported bitrate is %u (%u divider)\n",
			(unsigned int)bitrate, chip->bitrate, ssc_div);
#endif
	return 0;
}

static int snd_ssc_pcm_open_playback(struct snd_pcm_substream *substream)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int err;

	/* ensure buffer_size is a multiple of period_size */
	err = snd_pcm_hw_constraint_integer(runtime,
					SNDRV_PCM_HW_PARAM_PERIODS);

	if (err < 0)
		return err;

	runtime->hw = snd_ssc_playback_hw;
	chip->substream_playback = substream;

	return 0;
}

static int snd_ssc_pcm_close_playback(struct snd_pcm_substream *substream)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	chip->substream_playback = NULL;
	return 0;
}

static int snd_ssc_pcm_hw_params(struct snd_pcm_substream *substream,
				 struct snd_pcm_hw_params *hw_params)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	int channels = params_channels(hw_params);
	int val;

	unsigned int newrate = params_rate(hw_params);
	if( chip->bitrate != newrate )
	{
		if( snd_ssc_set_bitrate( chip, newrate ) < 0 )
			return -1;
		chip->bitrate = newrate;
	}
	
	val = ssc_readl(chip->ssc->regs, TFMR);
	val = SSC_BFINS(TFMR_DATNB, channels - 1, val);
	ssc_writel(chip->ssc->regs, TFMR, val);

	return snd_pcm_lib_malloc_pages(substream,
					params_buffer_bytes(hw_params));
}

static int snd_ssc_pcm_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

static int snd_ssc_pcm_prepare_playback(struct snd_pcm_substream *substream)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int block_size;

	block_size = frames_to_bytes(runtime, runtime->period_size);

	chip->period_playback = 0;

	ssc_writel(chip->ssc->regs, PDC_TPR,
			(long)runtime->dma_addr);
	ssc_writel(chip->ssc->regs, PDC_TCR,
			runtime->period_size * runtime->channels);
	ssc_writel(chip->ssc->regs, PDC_TNPR,
			(long)runtime->dma_addr + block_size);
	ssc_writel(chip->ssc->regs, PDC_TNCR,
			runtime->period_size * runtime->channels);

	return 0;
}

static int snd_ssc_pcm_trigger_playback(struct snd_pcm_substream *substream,
				   int cmd)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	int retval = 0;

	spin_lock(&chip->lock);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		ssc_writel(chip->ssc->regs, IER, SSC_BIT(IER_ENDTX));
		ssc_writel(chip->ssc->regs, PDC_PTCR, SSC_BIT(PDC_PTCR_TXTEN));
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		ssc_writel(chip->ssc->regs, PDC_PTCR, SSC_BIT(PDC_PTCR_TXTDIS));
		ssc_writel(chip->ssc->regs, IDR, SSC_BIT(IDR_ENDTX));
		break;
	default:
		//dev_dbg(&chip->spi->dev, "spurious command %x\n", cmd);
		printk(KERN_DEBUG "spurious command %x\n", cmd);
		retval = -EINVAL;
		break;
	}

	spin_unlock(&chip->lock);

	return retval;
}

static snd_pcm_uframes_t
snd_ssc_pcm_pointer_playback(struct snd_pcm_substream *substream)
{
	struct snd_ssc *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t pos;
	unsigned long bytes;

	bytes = ssc_readl(chip->ssc->regs, PDC_TPR)
		- (unsigned long)runtime->dma_addr;

	pos = bytes_to_frames(runtime, bytes);
	if (pos >= runtime->buffer_size)
		pos -= runtime->buffer_size;

	return pos;
}

static struct snd_pcm_ops ssc_playback_ops = {
	.open		= snd_ssc_pcm_open_playback,
	.close		= snd_ssc_pcm_close_playback,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= snd_ssc_pcm_hw_params,
	.hw_free	= snd_ssc_pcm_hw_free,
	.prepare	= snd_ssc_pcm_prepare_playback,
	.trigger	= snd_ssc_pcm_trigger_playback,
	.pointer	= snd_ssc_pcm_pointer_playback,
};

static void snd_ssc_pcm_free(struct snd_pcm *pcm)
{
	struct snd_ssc *chip = snd_pcm_chip(pcm);
	if (chip->pcm) {
		snd_pcm_lib_preallocate_free_for_all(chip->pcm);
		chip->pcm = NULL;
	}
}

static int __init snd_ssc_pcm_new(struct snd_ssc *chip, int device)
{
	struct snd_pcm *pcm;
	int retval = snd_pcm_new(chip->card, chip->card->shortname,
			device, 1, 0, &pcm);
	if (retval < 0)
		goto out;

	pcm->private_data = chip;
	pcm->private_free = snd_ssc_pcm_free;
	pcm->info_flags = SNDRV_PCM_INFO_BLOCK_TRANSFER;
	strcpy(pcm->name, DRV_NAME);
	chip->pcm = pcm;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &ssc_playback_ops);

	retval = snd_pcm_lib_preallocate_pages_for_all(chip->pcm,
			SNDRV_DMA_TYPE_DEV, &chip->ssc->pdev->dev,
			64 * 1024, 64 * 1024);
out:
	return retval;
}

static irqreturn_t snd_ssc_interrupt(int irq, void *dev_id)
{
	struct snd_ssc *chip = dev_id;
	struct snd_pcm_runtime *runtime;
	u32 status;
	int offset;
	int block_size;
	int next_period;
	int retval = IRQ_NONE;

	spin_lock(&chip->lock);

	status = ssc_readl(chip->ssc->regs, IMR);

	if (status & SSC_BIT(IMR_ENDTX)) {
		runtime = chip->substream_playback->runtime;
		chip->period_playback++;
		if (chip->period_playback == runtime->periods)
			chip->period_playback = 0;
		next_period = chip->period_playback + 1;
		if (next_period == runtime->periods)
			next_period = 0;

		block_size = frames_to_bytes(runtime, runtime->period_size);
		offset = block_size * next_period;

		ssc_writel(chip->ssc->regs, PDC_TNPR,
				(long)runtime->dma_addr + offset);
		ssc_writel(chip->ssc->regs, PDC_TNCR,
				runtime->period_size * runtime->channels);
		retval = IRQ_HANDLED;
		snd_pcm_period_elapsed(chip->substream_playback);
	}

	//ssc_readl(chip->ssc->regs, IMR);
	spin_unlock(&chip->lock);

	//if (status & SSC_BIT(IMR_ENDTX))
		//snd_pcm_period_elapsed(chip->substream_playback);

	return retval;
}

/*
 * Device functions
 */

static int __init snd_dac_chip_init(struct snd_ssc *chip)
{
	/* Enable I2S device, i.e. clock output. */
	ssc_writel(chip->ssc->regs, CR, SSC_BIT(CR_TXEN));

	return 0;
}

static int snd_ssc_dev_free(struct snd_device *device)
{
	struct snd_ssc *chip = device->device_data;

	ssc_writel(chip->ssc->regs, CR, SSC_BIT(CR_TXDIS) | SSC_BIT(CR_RXDIS));
	if (chip->irq >= 0)
	{
		free_irq(chip->irq, chip);
		chip->irq = -1;
	}

	return 0;
}

static int __init snd_ssc_init(struct snd_ssc *chip)
{
	/* Restart and disable SSC */
	ssc_writel(chip->ssc->regs, CR, SSC_BIT(CR_SWRST));
	ssc_writel(chip->ssc->regs, CMR, 0);
	ssc_writel(chip->ssc->regs, CR, SSC_BIT(CR_TXDIS) | SSC_BIT(CR_RXDIS));

	/*
	 * Continuous clock output.
	 * Starts on falling TF.
	 * Delay 1 cycle (1 bit).
	 * Periode is 16 bit (16 - 1).
	 */
	ssc_writel(chip->ssc->regs, TCMR,
			SSC_BF(TCMR_CKO, 1)
			| SSC_BF(TCMR_START, 4)
			| SSC_BF(TCMR_STTDLY, 1)
			| SSC_BF(TCMR_PERIOD, 16 - 1));

	/*
	 * Data length is 16 bit (16 - 1).
	 * Transmit MSB first.
	 * Transmit 2 words each transfer.
	 * Frame sync length is 16 bit (16 - 1).
	 * Frame starts on negative pulse.
	 */
	ssc_writel(chip->ssc->regs, TFMR,
			SSC_BF(TFMR_DATLEN, 16 - 1)
			| SSC_BIT(TFMR_MSBF)
			| SSC_BF(TFMR_DATNB, 1)
			| SSC_BF(TFMR_FSLEN, 16 - 1)
			| SSC_BF(TFMR_FSOS, 1));

	return 0;
}

static int __init snd_ssc_dev_init(struct snd_card *card)
{
	static struct snd_device_ops ops = {
		.dev_free	= snd_ssc_dev_free,
	};

	struct snd_ssc *chip = get_chip(card);
	int irq, retval;

	irq = chip->ssc->irq;
	if (irq < 0)
		return irq;

	spin_lock_init(&chip->lock);

	chip->card = card;
	chip->irq = -1;

	retval = request_irq(irq, snd_ssc_interrupt, 0, DRV_NAME, chip);
	if (retval) {
		printk("unable to request irq %d\n", irq);
		goto out;
	}
	chip->irq = irq;

	retval = snd_ssc_init(chip);
	if (retval)
	{
		printk("unable init ssc\n");
		goto out_irq;
	}

	retval = snd_dac_chip_init(chip);
	if (retval)
	{
		printk("unable init dac chip\n");
		goto out_irq;
	}

	retval = snd_ssc_pcm_new(chip, 0);
	if (retval)
	{
		printk("unable create new pcm\n");
		goto out_irq;
	}

	retval = snd_device_new(card, SNDRV_DEV_LOWLEVEL, chip, &ops);
	if (retval)
	{
		printk("unable create new snd device\n");
		goto out_irq;
	}

	goto out;

out_irq:
	free_irq(chip->irq, chip);
	chip->irq = -1;
out:
	return retval;
}

static int __init ssc_snd_init(void)
{
	struct snd_card			*card;
	struct snd_ssc		*chip;
	int				retval;
	char				id[16];

	/* Allocate "card" using some unused identifiers. */
	snprintf(id, sizeof id, DRV_NAME"_%d", SSC_ID);
	retval = snd_card_create(-1, id, THIS_MODULE,
				 sizeof(struct snd_ssc), &card);
	if (retval < 0)
		goto out;

	chip = (struct snd_ssc*)card->private_data;
	chip->bitrate = 0;

	chip->ssc = ssc_request(SSC_ID);
	if (IS_ERR(chip->ssc)) {
		printk("could not get ssc%d device\n",
				SSC_ID);
		retval = PTR_ERR(chip->ssc);
		goto out_card;
	}

	retval = snd_ssc_dev_init( card );
	if (retval)
		goto out_ssc;

	strcpy(card->driver, DRV_NAME);
	strcpy(card->shortname, "external sound on SSC");
	sprintf(card->longname, "%s on irq %d", card->shortname, chip->irq);

	retval = snd_card_register(card);
	if (retval)
		goto out_ssc;

	ssc_card = card;
	goto out;

out_ssc:
	ssc_free(chip->ssc);
out_card:
	snd_card_free(card);
out:
	return retval;
}

static void __exit ssc_snd_exit(void)
{
	struct snd_card *card = ssc_card;
	struct snd_ssc *chip = (struct snd_ssc*)card->private_data;

	/* Stop playback. */
	ssc_writel(chip->ssc->regs, CR, SSC_BIT(CR_TXDIS));
	ssc_free(chip->ssc);
	snd_card_free(card);
}

module_init(ssc_snd_init);
module_exit(ssc_snd_exit);

MODULE_DESCRIPTION("Sound driver for" DRV_NAME " with Atmel SSC");
MODULE_LICENSE("GPL");
